<?php

declare(strict_types=1);

namespace Drupal\batch\Batch;

use Drupal\awareness\Cache\MemoryCache\EntityMemoryCacheAwareTrait;
use Drupal\awareness\Entity\EntityTypeManagerAwareTrait;
use Drupal\Component\Utility\Bytes;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Base class for implementing batches.
 */
abstract class BatchBase extends BatchBuilder {

  use EntityMemoryCacheAwareTrait;
  use EntityTypeManagerAwareTrait;
  use LoggerChannelTrait;
  use StringTranslationTrait;

  /**
   * The number of items to process during on processing run.
   *
   * @var int
   */
  protected int $itemsPerProcess = 25;

  /**
   * URL for redirection upon batch completion.
   *
   * @var \Drupal\Core\Url|null
   */
  protected Url|null $redirectUrl = NULL;

  /**
   * {@inheritdoc}
   */
  public function __construct() {
    parent::__construct();
    $this->addOperation([$this, 'process']);
    $this->setFinishCallback([$this, 'finished']);
  }

  /**
   * Batch operation callback.
   *
   * @param array $context
   *   The batch context.
   */
  abstract public function process(&$context);

  /**
   * Batch finished callback.
   *
   * @param bool $success
   *   A boolean indicating whether the batch has completed successfully.
   * @param mixed $results
   *   The value set in $context['results'] by callback_batch_operation().
   * @param array $operations
   *   If $success is FALSE, contains the operations that remained unprocessed.
   */
  public function finished(bool $success, mixed $results, array $operations): RedirectResponse|null {
    // Override as needed for message setting, etc. and be sure to utilize this
    // return value to keep the redirection functionality.
    if (!empty($this->redirectUrl)) {
      return new RedirectResponse($this->redirectUrl->toString());
    }
  }

  /**
   * Reclaims memory.
   *
   * @return bool
   *   Indicates if memory was reclaimed.
   *
   * @see \Drupal\migrate\MigrateExecutable::attemptMemoryReclaim
   */
  protected function reclaimMemory() {
    $limit = trim(ini_get('memory_limit'));
    if ($limit == '-1') {
      $limit = Bytes::toNumber('256MB');
    }
    else {
      $limit = Bytes::toNumber($limit);
    }
    if (memory_get_usage() / $limit < .85) {
      return FALSE;
    }

    drupal_static_reset();

    // Entity storage can blow up with caches so clear them out.
    foreach ($this->getEntityTypeManager()->getDefinitions() as $id => $definition) {
      $this->getEntityTypeManager()->getStorage($id)->resetCache();
    }

    // Clear the entity storage memory cache.
    /** @var \Drupal\Core\Cache\MemoryCache\MemoryCache $memory_cache */
    $this->getEntityMemoryCache()->deleteAll();

    // Run garbage collector to further reduce memory.
    gc_collect_cycles();
    return TRUE;
  }

  /**
   * Set the number of items to process during each processing run.
   *
   * @param int $itemsPerProcess
   *   The number of items to process during each processing run.
   *
   * @return $this
   */
  public function setItemsPerProcess($itemsPerProcess) {
    $this->itemsPerProcess = $itemsPerProcess;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function setRedirectUrl(Url $url) {
    $this->redirectUrl = $url;
    return $this;
  }

}
